import gymnasium as gym 
import numpy as np
import warnings
import torch
import sys 
import os 
sys.path.insert(1, os.path.join(sys.path[0], '..'))

from src.graph_envs import shortest_path, utils

env_id = 'MaxIndependentSet-v0'
kwargs = {'n_nodes': 5, 'n_edges': 7, 'return_graph_obs': True, 'weighted': False}

def test_max_independent_set():
    
    env = gym.make(env_id, **kwargs)
    
    for _ in range(5):
        
        obs, info = env.reset()
        
        mask = info['mask']
        done = False
        graph_obs = info['graph_obs']

        obs = torch.Tensor(obs.reshape(1, -1))
        nodes, edges, edge_links = utils.devectorize_graph(vector=obs, env_id=env_id, **kwargs)
        
        assert (nodes == torch.Tensor(graph_obs.nodes)).all(), f"{obs.nodes} {graph_obs.nodes}"
        assert (edges == torch.Tensor(graph_obs.edges)).all(), f"{obs.edges} {graph_obs.edges}"
        assert (edge_links == torch.Tensor(graph_obs.edge_links)).all(), f"{obs.edge_links} {graph_obs.edge_links}"
        assert (nodes[:, :, 1]).abs().sum() == 0, nodes[:, :, 1]
        
        
        while not done:
            valid_actions = mask.nonzero()[0]
            action = np.random.choice(valid_actions)
            obs, reward, done, _, info = env.step(action)
            nodes, edges, edge_links = utils.devectorize_graph(vector=torch.Tensor(obs.reshape(1, -1)), env_id=env_id, **kwargs)
            mask = info['mask']
            
    assert True
    
if __name__ == '__main__':
    test_max_independent_set()



import gymnasium as gym
from gymnasium.envs.registration import register

name = "graph_envs"

## For local testing and running:

# register(
#     id='ShortestPath-v0',
#     entry_point='src.graph_envs.shortest_path:ShortestPathEnv',
# )

# register(
#     id='SteinerTree-v0',
#     entry_point='src.graph_envs.steiner_tree:SteinerTreeEnv',
# )

# register(
#     id='MaxIndependentSet-v0',
#     entry_point='src.graph_envs.max_independent_set:MaxIndependentSet',
# )

# register(
#     id='TSP-v0',
#     entry_point='src.graph_envs.tsp:TSPEnv',
# )


## For packaging:

register(
    id='ShortestPath-v0',
    entry_point='graph_envs.shortest_path:ShortestPathEnv',
)

register(
    id='SteinerTree-v0',
    entry_point='graph_envs.steiner_tree:SteinerTreeEnv',
)

register(
    id='MaxIndependentSet-v0',
    entry_point='graph_envs.max_independent_set:MaxIndependentSet',
)

register(
    id='TSP-v0',
    entry_point='graph_envs.tsp:TSPEnv',
)


import gymnasium as gym
# import torch 
# import torch_geometric as pyg 
import numpy as np 

from typing import TYPE_CHECKING, Any, Generic, SupportsFloat, TypeVar

import networkx as nx 
import random

from typing import Tuple


def get_prime_numbers(n):
    '''
    Returns a list of prime numbers less than n
    '''
    prime_numbers = []
    for num in range(2, n):
        if all(num % i != 0 for i in range(2, num)):
            prime_numbers.append(num)
    return prime_numbers

class SteinerTreeEnv(gym.Env):
    '''
    Environment for shortest path problem.
    input:
        - n_nodes: number of nodes in the graph
        - n_edges: number of edges in the graph
        - weighted: whether the graph is weighted or not
        - n_dests: number of destinations to be reached
    '''
    
    def __init__(self, n_nodes, n_edges, n_dests=3, weighted=True, is_eval_env=False) -> None:
        super(SteinerTreeEnv, self).__init__()
        
        # Node status codes
        # self.HAS_NOTHING = np.array([0.0], dtype=np.float32)
        # self.HAS_MSG = np.array([2.0], dtype=np.float32)
        # self.IS_TARGET = np.array([3.0], dtype=np.float32)
        self.NODE_HAS_MSG = 0
        self.NODE_IS_TARGET = 1
        
        # Edge status codes
        # self.IS_NOT_TAKEN = np.array([1.0], dtype=np.float32)
        # self.HAS_MSG = np.array([2.0], dtype=np.float32)
        self.EDGE_IS_TAKEN = 1
        self.EDGE_WEIGHT = 0
        
        self.n_nodes = n_nodes
        self.n_edges = n_edges
        self.n_dests = n_dests
        self.weighted = weighted
        self.action_space = gym.spaces.Discrete(n_edges)
        # self.observation_space = gym.spaces.Box(low=0, high=1000, shape=(n_nodes+2*n_edges+2*n_edges*2,))
        self.observation_space = gym.spaces.Box(low=0, high=1000, shape=(2*n_nodes+2*n_edges*2+2*n_edges*2,))
        # self.observation_space = gym.spaces.Graph(
        #     node_space=gym.spaces.Box(low=0, high=4, shape=(1,)), 
        #     edge_space=gym.spaces.Box(low=0, high=1, shape=(1,)),
        #     )
        self.is_eval_env = is_eval_env
        

    def reset(self, seed=None, options={}) -> np.array:
        super().reset(seed=seed)
        random.seed(seed)
        np.random.seed(seed)
        
        while True:
            G = nx.gnm_random_graph(self.n_nodes, self.n_edges)
            if nx.is_connected(G):
                break
        
        if self.weighted:
            delay = np.random.randint(3, 10, size=(self.n_nodes, self.n_nodes))/10.0
        else:
            delay = np.random.randint(10, 11, size=(self.n_nodes, self.n_nodes))/10.0
        
        for u, v, d in G.edges(data=True):
            d['delay'] = delay[u, v]
        
       
        # x = np.zeros((self.n_nodes, 1), dtype=np.float32) + self.HAS_NOTHING
        x = np.zeros((self.n_nodes, 2), dtype=np.float32)      
        
        self.dests = np.random.choice(self.n_nodes, self.n_dests+1, replace=False)
        
        self.approx_solution = 0
        
        if self.is_eval_env:
            if self.n_dests == 1:
                self.approx_solution = nx.algorithms.shortest_paths.generic.shortest_path_length(G, self.dests[0], self.dests[1], weight='delay')
            elif self.n_dests == self.n_nodes - 1:
                self.approx_solution = sum([features['delay'] for _, _, features in nx.minimum_spanning_edges(G, weight='delay')])
                approx_solution_graph = nx.algorithms.approximation.steinertree.steiner_tree(G, self.dests, weight='delay', method='kou')  
            else:
                approx_solution_graph = nx.algorithms.approximation.steinertree.steiner_tree(G, self.dests, weight='delay', method='kou')  
                self.approx_solution = sum([G[u][v]['delay'] for u, v in approx_solution_graph.edges()])
            
        G = G.to_directed()
        
        self.src, self.dests = self.dests[0], self.dests[1:]
        
        
        x[self.src, self.NODE_HAS_MSG] = 1
        x[self.dests, self.NODE_IS_TARGET] = 1
        
        self.adj = nx.adjacency_matrix(G).todense()
        
        edge_index = np.array(list(G.edges))
        edge_f = np.array([G[u][v]['delay'] for u, v in G.edges], dtype=np.float32).reshape(-1, 1)
       
        edge_f = np.concatenate((edge_f, np.zeros((2*self.n_edges, 1), dtype=np.float32)), axis=1)
        
        
        self.graph = gym.spaces.GraphInstance(nodes=x, edges=edge_f, edge_links=edge_index)
        self.generated_solution = []

        self.solution_cost = 0
        info = {'mask': self._get_mask()}
        self._vectorize_graph(self.graph)
        return self._vectorize_graph(self.graph), info
    
    
    def _vectorize_graph(self, graph):
        return np.concatenate((graph.nodes.flatten(), graph.edges.flatten(), graph.edge_links.flatten()), dtype=np.float32)
    
    def _get_mask(self) -> np.array:
        mask = np.zeros((2 * self.n_edges,), dtype=bool) < 1.0
        mask[self.graph.nodes[self.graph.edge_links[:, 0], self.NODE_HAS_MSG] < 0.5] = False
        mask[self.graph.nodes[self.graph.edge_links[:, 1], self.NODE_HAS_MSG] > 0.5] = False
        return mask
    
    
    def step(self, action: Tuple[int, int]) -> Tuple[gym.spaces.GraphInstance, SupportsFloat, bool, bool, dict]:
        
        
        u, v = self.graph.edge_links[action, 0], self.graph.edge_links[action, 1]
        
        # print(self.graph.nodes.T)
        assert (u < self.n_nodes), f"Node {u} is out of bounds!"
        assert (v < self.n_nodes), f"Node {v} is out of bounds!"
        assert self._get_mask()[action] == True, f"Mask of {action} is False!"
        assert np.isclose(self.graph.edges[action, self.EDGE_IS_TAKEN], 1) == False, f"BUG IN THE LIBRARY! Edge {action} is already a part of the path!"
        assert np.isclose(self.graph.nodes[u, self.NODE_HAS_MSG], 1), f"BUG IN THE LIBRARY! Node {u} is not a part of the path!"
        assert np.isclose(self.graph.nodes[v, self.NODE_HAS_MSG], 0), f"BUG IN THE LIBRARY! Node {v} is already a part of the path!"
        
        done = False
        reward = -self.graph.edges[action, self.EDGE_WEIGHT]
        self.solution_cost -= reward
        self.graph.nodes[v, self.NODE_HAS_MSG] = 1
        self.generated_solution.append((u, v))
        
        
        # print(self.graph.nodes[:, self.NODE_IS_TARGET] == 1)
        # print(self.graph.nodes[:, self.NODE_HAS_MSG] == 0)
        # print(np.logical_and(self.graph.nodes[:, self.NODE_HAS_MSG] == 0, self.graph.nodes[:, self.NODE_IS_TARGET] == 1))
         
        if np.logical_and(self.graph.nodes[:, self.NODE_HAS_MSG] == 0, self.graph.nodes[:, self.NODE_IS_TARGET] == 1).sum() == 0:
            done = True
   
            
        info = {'mask': self._get_mask()}
        
        if done:
            info['heuristic_solution'] = self.approx_solution
            info['solved'] = True
            info['solution_cost'] = self.solution_cost
            pass
        else:
            assert info['mask'].sum() > 0, "No more actions possible! Shouldn't happen!"
        
        
        return self._vectorize_graph(self.graph), reward, done, False, info
        


import gymnasium as gym 
from src.graph_envs import shortest_path, utils
import numpy as np
import warnings

env_id = 'ShortestPath-v0'
kwargs = {'n_nodes': 5, 'n_edges': 7, 'return_graph_obs': True}

def test_shortest_path():
    
    env = gym.make(env_id, **kwargs)

    for _ in range(5):
        
        obs, info = env.reset()
        mask = info['mask']
        done = False
        graph_obs = info['graph_obs']
    
        obs = utils.devectorize_graph(vector=obs, env_id=env_id, **kwargs)
        
        assert (obs.nodes == graph_obs.nodes).all(), f"{obs.nodes} {graph_obs.nodes}"
        assert (obs.edges == graph_obs.edges).all(), f"{obs.edges} {graph_obs.edges}"
        assert (obs.edge_links == graph_obs.edge_links).all(), f"{obs.edge_links} {graph_obs.edge_links}"
        assert (obs.nodes == env.IS_TARGET).sum() == 1
        
        while not done:
            valid_actions = mask.nonzero()[0]
            action = np.random.choice(valid_actions)
            obs, reward, done, _, info = env.step(action)
            obs = utils.devectorize_graph(vector=obs, env_id=env_id, **kwargs)
            assert (obs.nodes == graph_obs.nodes).all(), f"{obs.nodes} {graph_obs.nodes}"
            assert (obs.edges == graph_obs.edges).all(), f"{obs.edges} {graph_obs.edges}"
            assert (obs.edge_links == graph_obs.edge_links).all(), f"{obs.edge_links} {graph_obs.edge_links}"
        
            mask = info['mask']
    assert True


import gymnasium as gym
import torch 
import torch_geometric as pyg 
import numpy as np 

import networkx as nx 
import random
from typing import Tuple

# from MILP import flow 

def sample_graph(N, E):
    G = nx.random_tree(N)
    while G.number_of_edges() < E:
        x, y = random.randint(0, N-1), random.randint(0, N-1)
        if x == y:
            continue
        if (x, y) in G.edges:
            continue
        G.add_edge(x, y)
        
    assert nx.is_connected(G)
    return G

class MultiCastEnv(gym.Env):
    def __init__(self, n_nodes, n_edges, n_targets, n_msgs, is_eval_env=False) -> None:
        super(MultiCastEnv, self).__init__()
        
        self.HAS_NOTHING = np.array([1.0])
        self.HAS_MSG = np.array([2.0])
        self.IS_TARGET = np.array([3.0])
        self.inf_delay = np.array([-1.0])
        
        self.n_nodes = n_nodes
        self.n_edges = n_edges
        self.n_targets = n_targets
        self.n_msgs = n_msgs
        
        self.action_space = gym.spaces.MultiDiscrete([n_nodes, n_nodes])
        self.observation_space = gym.spaces.Box(low=-100., high=100., shape=(n_msgs+2, n_nodes, n_nodes))        
        self.is_eval_env = is_eval_env
        
    def reset(self, seed=None, options={}) -> np.array:
        super().reset(seed=seed)
        random.seed(seed)
        np.random.seed(seed)
        
        G = sample_graph(self.n_nodes, self.n_edges)
        
        bandwidth = np.random.randint(1,2, size=(self.n_nodes, self.n_nodes))
        delay = np.random.randint(2,5, size=(self.n_nodes, self.n_nodes))/1.0
        
        for u, v, d in G.edges(data=True):
            d['delay'] = delay[u, v]
            d['bandwidth'] = bandwidth[u, v]
        
        selected_nodes = np.random.choice(self.n_nodes, self.n_msgs*(1 + self.n_targets), replace=False)
        
        ''' Single Graph case for debug '''
        # G = G_global.copy()
        # selected_nodes = selected_nodes_global.copy()
        
        # asp = nx.floyd_warshall_numpy(G, delay='delay')
        adj = np.array(nx.adjacency_matrix(G, dtype=float, weight='delay').todense(), dtype=np.float32)
        adj[adj==0] = self.inf_delay
        np.fill_diagonal(adj, 0)
        
        bw = np.array(nx.adjacency_matrix(G, dtype=float, weight='bandwidth').todense(), dtype=np.float32)
        bw[bw==0] = self.inf_delay
        np.fill_diagonal(bw, 0)
        
        state = np.zeros((self.n_msgs, self.n_nodes, self.n_nodes)) + self.HAS_NOTHING
        
        msg_info = {
            i: {
                'sources': [selected_nodes[i*(1 + self.n_targets)]],
                'destinations': selected_nodes[i*(1 + self.n_targets) + 1: i*(1 + self.n_targets) + 1 + self.n_targets],
                'size': 1
            } for i in range(self.n_msgs)}

        
        for msg in range(self.n_msgs):
            pad = msg * (1 + self.n_targets)
            state[msg, selected_nodes[pad + 0],:] = self.HAS_MSG
            state[msg, selected_nodes[pad + 1: pad + self.n_targets + 1],:] = self.IS_TARGET

        
        # self.x = np.stack([np.zeros_like(adj) + self.HAS_NOTHING, self.x], axis=1, dtype=np.float32)
        self.x = np.concatenate((state, np.expand_dims(adj, 0) , np.expand_dims(bw, 0)), axis=0, dtype=np.float32)
        
        
        self.edge_index = pyg.utils.from_networkx(G).edge_index
        
        info = {'mask': self.get_mask()}
        return self.x, info
    
    def get_mask(self) -> np.array:
        
        valid_edges = np.logical_and(
            np.isclose(self.x[0:self.n_msgs, self.edge_index[0], 0], self.HAS_MSG),
            np.isclose(self.x[0:self.n_msgs, self.edge_index[1], 0], self.HAS_MSG) == False
            ) # shape: (2*n_edges,)
        
        mask = np.zeros((self.n_nodes, self.n_msgs * self.n_nodes), dtype=bool)
        bw_mask = (self.x[-1,:,:] > 0)
        
        for msg in range(self.n_msgs):
            pad = msg*self.n_nodes
            mask[self.edge_index[0, valid_edges[msg]], pad+self.edge_index[1, valid_edges[msg]]] = True
            mask[:, pad:pad+self.n_nodes] = np.logical_and(mask[:, pad:pad+self.n_nodes], bw_mask)
        
        return mask
    
    def step(self, action: Tuple[int, int, int]) -> Tuple[np.array, bool, dict]:
        
        u, v, msg = action
        
        assert (u < self.n_nodes) and (v < self.n_nodes), f"{u, v} is not a valid edge!"
        assert self.get_mask()[u, msg*self.n_nodes+v] == True, f"Mask of {u, v} is False!"
        assert not np.isclose(self.x[-2, u, v], self.inf_delay), f"{u, v} is not an edge!"
        assert np.isclose(self.x[msg, u, 0], self.HAS_MSG), f"{u} is not a source!"
        assert np.isclose(self.x[msg, v, 0], self.HAS_MSG) == False, f"{v} is a destination!"
        assert not np.isclose(self.x[-1, u, v], 0), f"{u, v}'s bandwidth is not enough!"
        
        done = False
        reward = -self.x[-2, u, v]
        
        if self.is_eval_env:
            self.solution += self.x[-2, u, v]
        
        if np.isclose(self.x[msg, v, 0], self.IS_TARGET):
            reward += 1
            
        self.x[msg, v, :] = self.HAS_MSG
        self.x[-1, u, v] -= 1
        
        
        if not np.isclose(self.x[0:self.n_msgs, :, 0], self.IS_TARGET).any():
            done = True
        elif (self.get_mask().sum() == 0):
            done = True
            reward = -10
            
        info = {'mask': self.get_mask()}
        if self.is_eval_env:
            info['opt'] = self.opt
            info['sol'] = self.solution

            
        return self.x, reward, done, False, info
        


import numpy as np 
import gymnasium as gym 
import torch 



def devectorize_graph(vector, env_id, **kwargs):
    bs = vector.shape[0]
    
    if env_id == "ShortestPath-v0":
        x = vector[:, :2*kwargs['n_nodes']].reshape(bs, kwargs['n_nodes'], 2)
        edge_features = vector[:, 2*kwargs['n_nodes']:2*kwargs['n_nodes'] + 2*kwargs['n_edges']].reshape(bs, 2*kwargs['n_edges'], 1)
        edge_index = vector[:, 2*kwargs['n_nodes'] + 2*kwargs['n_edges']:].reshape(bs, 2*kwargs['n_edges'], 2).long()
    elif env_id == "SteinerTree-v0":
        x = vector[:, :2*kwargs['n_nodes']].reshape(bs, kwargs['n_nodes'], 2)
        edge_features = vector[:, 2*kwargs['n_nodes']: 2*kwargs['n_nodes']+2*kwargs['n_edges']*2].reshape(bs, 2*kwargs['n_edges'], 2)
        edge_index = vector[:, 2*kwargs['n_nodes'] + 2*kwargs['n_edges']*2:].reshape(bs, 2*kwargs['n_edges'], 2).long()
    elif env_id == "MaxIndependentSet-v0":
        x = vector[:, :2*kwargs['n_nodes']].reshape(bs, kwargs['n_nodes'], 2)
        edge_features = vector[:, 2*kwargs['n_nodes']: 2*kwargs['n_nodes']+2*kwargs['n_edges']].reshape(bs, 2*kwargs['n_edges'], 1)
        edge_index = vector[:, 2*kwargs['n_nodes'] + 2*kwargs['n_edges']:].reshape(bs, 2*kwargs['n_edges'], 2).long()
    elif env_id == "TSP-v0":
        x = vector[:, :kwargs['n_nodes']].reshape(bs, kwargs['n_nodes'], 1)
        edge_features = vector[:, kwargs['n_nodes']:kwargs['n_nodes'] + 2*kwargs['n_edges']].reshape(bs, 2*kwargs['n_edges'], 1)
        edge_index = vector[:, kwargs['n_nodes'] + 2*kwargs['n_edges']:].reshape(bs, 2*kwargs['n_edges'], 2).long()
    return x, edge_features, edge_index
    

def to_pyg_graph(x, edge_features, edge_index):
    import torch_geometric as pyg
    graphs = [pyg.data.Data(x=x[i,:,:], edge_attr=edge_features[i,:,:], edge_index=edge_index[i,:,:].T) for i in range(x.shape[0])]
    batch = pyg.data.Batch.from_data_list(graphs).to(x.device)
    return batch




import gymnasium as gym 
from src.graph_envs import shortest_path
import numpy as np


def test_shortest_path():
    for tst in range(1, 5):
        
        n_dests = np.random.randint(1, 5)
        
        env = gym.make('SteinerTree-v0', n_nodes=5, n_edges=10, n_dests=tst)

        for _ in range(5):
            print('---------')
            
            obs, info = env.reset()
            mask = info['mask']
            print(obs)
            
            done = False
            assert (obs.nodes == env.IS_TARGET).sum() == tst
            
            while not done:
                valid_actions = mask.nonzero()[0]
                action = np.random.choice(valid_actions)
                print(action)
                obs, reward, done, _, info = env.step(action)
                mask = info['mask']
            print(info['generated_solution'])
        assert True
        
        
if __name__ == '__main__':
    test_shortest_path()


import gymnasium as gym
# import torch 
# import torch_geometric as pyg 
import numpy as np 

from typing import TYPE_CHECKING, Any, Generic, SupportsFloat, TypeVar, Tuple

import networkx as nx 
import random


class ShortestPathEnv(gym.Env):
    '''
    Environment for shortest path problem.
    input:
        - n_nodes: number of nodes in the graph
        - n_edges: number of edges in the graph
        - weighted: whether the graph is weighted or not
    '''
    
    def __init__(self, n_nodes, n_edges, weighted=True, return_graph_obs=False, is_eval_env=False) -> None:
        super(ShortestPathEnv, self).__init__()
        
        self.HAS_NOTHING = np.array([0.0], dtype=np.float32)
        self.IS_TAKEN = np.array([1.0], dtype=np.float32)
        self.IS_TARGET = np.array([1.0], dtype=np.float32)
        
        self.n_nodes = n_nodes
        self.n_edges = n_edges
        self.weighted = weighted
        self.action_space = gym.spaces.Discrete(n_nodes)
        # self.observation_space = gym.spaces.Box(low=-1., high=11., shape=(2, n_nodes, n_nodes))        
        # self.observation_space = gym.spaces.Graph(
        #     node_space=gym.spaces.Box(low=0, high=4, shape=(1,)), 
        #     edge_space=gym.spaces.Box(low=0, high=1, shape=(1,)),
        #     )
        self.observation_space = gym.spaces.Box(low=0, high=1000, shape=(2*n_nodes+2*n_edges+2*n_edges*2,))
        self.return_graph_obs = return_graph_obs
        self.solution_cost = 0
        self.is_eval_env = is_eval_env
        
    def reset(self, seed=None, options={}) -> np.array:
        super().reset(seed=seed)
        random.seed(seed)
        np.random.seed(seed)
        
        while True:
            G = nx.gnm_random_graph(self.n_nodes, self.n_edges)
            if nx.is_connected(G):
                break
        
        if self.weighted:
            delay = np.random.randint(3, 10, size=(self.n_nodes, self.n_nodes))/10.0
            # delay = np.random.randint(1, 3, size=(self.n_nodes, self.n_nodes))*1.0
        else:
            delay = np.random.randint(10, 11, size=(self.n_nodes, self.n_nodes))/10.0
            
        
        for u, v, d in G.edges(data=True):
            d['delay'] = delay[u, v]
        
        G = G.to_directed()
        
        x = np.zeros((self.n_nodes, 2), dtype=np.float32) + self.HAS_NOTHING
        
        self.src, self.dest = np.random.choice(self.n_nodes, size=2, replace=False)
        x[self.src, 0], x[self.dest, 1] = self.IS_TAKEN, self.IS_TARGET
        self.head = self.src
        self.adj = nx.adjacency_matrix(G, weight='delay').todense()
        
        edge_index = np.array(list(G.edges))
        edge_f = np.array([G[u][v]['delay'] for u, v in G.edges], dtype=np.float32).reshape(-1, 1)
        self.graph = gym.spaces.GraphInstance(nodes=x, edges=edge_f, edge_links=edge_index)
        
        self.optimal_solution = 0
        if self.is_eval_env:
            self.optimal_solution = nx.shortest_path_length(G, source=self.src, target=self.dest, weight='delay')
        
        
        # print('-----')
        # print(self.graph.nodes)
        info = {'mask': self._get_mask()}
        
        if self.return_graph_obs:
            info['graph_obs'] = self.graph
        
        self.solution_cost = 0
        return self._vectorize_graph(self.graph), info
    
    def _vectorize_graph(self, graph):
        return np.concatenate((graph.nodes.flatten(), graph.edges.flatten(), graph.edge_links.flatten()), dtype=np.float32)
    
    
    def _get_neighbors(self, node):
        neigbors = self.graph.edge_links[self.graph.edge_links[:, 0] == node, 1]
        return neigbors
        
    def _get_mask(self) -> np.array:
        mask = np.zeros((self.n_nodes,), dtype=bool)
        mask[self._get_neighbors(self.head)] = 1
        mask[self.graph.nodes[:, 0] == self.IS_TAKEN] = 0
        return mask
    
    def step(self, action: int) -> Tuple[gym.spaces.GraphInstance, SupportsFloat, bool, bool, dict]:
        assert (action < self.n_nodes), f"Node {action} is out of bounds!"
        assert self._get_mask()[action] == True, f"Mask of {action} is False!"

        assert action in self._get_neighbors(self.head), f"Node {action} is not a neighbor of the current path head {self.head}!"
        assert np.isclose(self.graph.nodes[action, 0], self.IS_TAKEN) == False, f"Node {action} is already a part of the path!"
        
        
        
        done = False
        reward = -self.adj[self.head, action]
        self.solution_cost -= reward
        info = {}
        
        if np.isclose(self.graph.nodes[action, 1], self.IS_TARGET):
            # reward += self.n_nodes
            done = True
            info['solved'] = True
        
        
        self.graph.nodes[action, 0] = self.IS_TAKEN
        self.head = action
        
        info['mask'] = self._get_mask()
        if (not done) and (info['mask'].sum() == 0):
            done = True
            reward = -self.n_nodes
            info['solved'] = False
            
        if done:
            info['heuristic_solution'] = self.optimal_solution
            info['solution_cost'] = self.solution_cost
        return self._vectorize_graph(self.graph), reward, done, False, info
        



import torch 
import torch.nn as nn
import gymnasium as gym
import numpy as np 
import random 
import networkx as nx

from typing import Tuple, SupportsFloat

class MaxIndependentSet(gym.Env):
    '''
    Environment for shortest path problem.
    input:
        - n_nodes: number of nodes in the graph
        - n_edges: number of edges in the graph
        - weighted: whether the nodes are weighted or not
    '''
    
    def __init__(self, n_nodes, n_edges, weighted=True, return_graph_obs=False, is_eval_env=False) -> None:
        super(MaxIndependentSet, self).__init__()
        
        
        self.NODE_WEIGHT = 0
        self.NODE_IS_TAKEN = 1
        
        self.n_nodes = n_nodes
        self.n_edges = n_edges
        self.weighted = weighted
        self.action_space = gym.spaces.Discrete(n_nodes)
        
        self.return_graph_obs = return_graph_obs
        self.observation_space = gym.spaces.Box(low=0, high=1000, shape=(2*n_nodes+2*n_edges+2*n_edges*2,))
        self.is_eval_env = is_eval_env
        
        
    def reset(self, seed=None, options={}) -> np.array:
        super().reset(seed=seed)
        random.seed(seed)
        np.random.seed(seed)
        
        while True:
            G = nx.gnm_random_graph(self.n_nodes, self.n_edges)
            if nx.is_connected(G):
                break
        
        if self.weighted:
            cost = np.random.randint(3, 10, size=(self.n_nodes))/10.0
            
        else:
            cost = np.random.randint(10, 11, size=(self.n_nodes))/10.0
            
        for u in G.nodes():
            G.nodes[u]['cost'] = cost[u]
            
        self.approx_solution = 0
        if self.is_eval_env:
            if not self.weighted:
                self.approx_solution = len(nx.approximation.maximum_independent_set(G))
            else:
                self.approx_solution = -1
            
        G = G.to_directed()
        
        x = np.zeros((self.n_nodes, 2), dtype=np.float32)
        x[:, 0] = cost
        
        edge_index = np.array(list(G.edges))
        edge_f = np.array([1.0 for u, v in G.edges], dtype=np.float32).reshape(-1, 1)
        
        self.graph = gym.spaces.GraphInstance(nodes=x, edges=edge_f, edge_links=edge_index)
        
        info = {'mask': self._get_mask()}
        
        if self.return_graph_obs:
            info['graph_obs'] = self.graph
        
        self.solution_cost = 0
        return self._vectorize_graph(self.graph), info
    
    
    def _vectorize_graph(self, graph):
        return np.concatenate((graph.nodes.flatten(), graph.edges.flatten(), graph.edge_links.flatten()), dtype=np.float32)
    
    def _get_mask(self) -> np.array:
        
        # self.graph.nodes[3,1] = 1
        # self.graph.nodes[1,1] = 1
        
        mask = np.zeros((self.n_nodes,), dtype=bool)
        mask[self.graph.nodes[:, self.NODE_IS_TAKEN] == 0] = True
        
        edges_with_taken_start = self.graph.edge_links[self.graph.nodes[self.graph.edge_links[:, 0], self.NODE_IS_TAKEN] == 1]
        their_neighbours = edges_with_taken_start[:, 1]
        mask[their_neighbours] = False
        return mask

    def step(self, action: int) -> Tuple[gym.spaces.GraphInstance, SupportsFloat, bool, bool, dict]:
        assert (action < self.n_nodes), f"Node {action} is out of bounds!"
        assert self._get_mask()[action] == True, f"Mask of {action} is False!"

        assert np.isclose(self.graph.nodes[action, self.NODE_IS_TAKEN], 0), f"Node {action} is already a part of the path!"
        
        done = False
        reward = -self.graph.nodes[action, self.NODE_WEIGHT]
        self.solution_cost -= reward
        info = {}
        
        self.graph.nodes[action, self.NODE_IS_TAKEN] = 1
        info['mask'] = self._get_mask()
        
        if info['mask'].sum() == 0:
            done = True
            info['solved'] = True      
        
        if done:
            info['heuristic_solution'] = self.approx_solution
            info['solution_cost'] = self.solution_cost
        
        return self._vectorize_graph(self.graph), reward, done, False, info
          
        
        
        
            